from .atmosphere import AtmosphericForcingEntry
from .hydrology import NationalWaterModelEntry
from .ice import IceForcingEntry
from .ocean import ADCIRCEntry, SCHISMEntry
from .waves import SWANEntry, WaveWatch3Entry, WaveWatch3ForcingEntry



from datetime import datetime, timedelta
from os import PathLike
from pathlib import Path
from typing import Dict, List

from .configuration import (
    ConfigurationFile,
    ensure_directory,
    FileForcingsFile,
    ModelConfigurationFile,
    NEMSConfigurationFile,
    RunSequence,
)
from .model.base import ConnectionEntry, EntryType, GridRemapMethod, MediationEntry, ModelEntry
from .utilities import parse_datetime


class ModelingSystem:
    """
    main user interface class of NEMSpy, providing configuration and output to files

    >>> from datetime import datetime, timedelta
    >>> from nemspy import ModelingSystem
    >>> from nemspy.model import ADCIRCEntry, AtmosphericForcingEntry, WaveWatch3ForcingEntry
    >>> # instantiate model system with model entries
    >>> nems = ModelingSystem(
    >>>     start_time=datetime(2020, 6, 1),
    >>>     end_time=start_time + timedelta(days=1),
    >>>     interval=timedelta(hours=1),
    >>>     ocn=ADCIRCEntry(processors=11),
    >>>     atm=AtmosphericForcingEntry(filename='wind_atm_fin_ch_time_vec.nc'),
    >>>     wav=WaveWatch3ForcingEntry(filename='ww3.Constant.20151214_sxy_ike_date.nc'),
    >>> )
    >>> # form connections between models
    >>> nems.connect('ATM', 'OCN')
    >>> nems.connect('WAV', 'OCN')
    >>> # define execution order
    >>> nems.sequence = [
    >>>     'ATM -> OCN',
    >>>     'WAV -> OCN',
    >>>     'ATM',
    >>>     'WAV',
    >>>     'OCN',
    >>> ]
    >>> # write configuration files to the given directory
    >>> nems.write('nems_configuration')
    """

    def __init__(
        self, start_time: datetime, end_time: datetime, interval: timedelta, **models,
    ):
        """
        :param start_time: start time within the modeled system
        :param end_time: end time within the modeled system
        :param interval: time interval of top-level run sequence
        :param verbose: verbosity in NEMS configuration
        :param atm: atmospheric wind model
        :param wav: oceanic wave model
        :param ocn: oceanic circulation model
        :param hyd: terrestrial water model
        :param med: model mediator
        """

        self.__start_time = start_time
        self.end_time = end_time

        model_types = [model_type.value.lower() for model_type in EntryType]

        parsed_models = {}
        attributes = {}
        for key, value in models.items():
            if key.lower() in model_types:
                key = key.lower()
                if isinstance(value, ModelEntry):
                    if value.entry_type.value.lower() == key:
                        parsed_models[key.upper()] = value
                    else:
                        raise TypeError(f'"{value.name}" is not {key}')
                else:
                    raise TypeError(f'unsupported type {value.__class__}"')
            else:
                attributes[key] = value

        models = parsed_models
        self.__sequence = RunSequence(interval, **models, **attributes)

    @property
    def start_time(self) -> datetime:
        """
        end time within modeled system
        """

        return self.__start_time

    @start_time.setter
    def start_time(self, start_time: datetime):
        start_time = parse_datetime(start_time)
        self.__start_time = start_time
        if self.start_time > self.end_time:
            self.start_time = self.end_time
            self.end_time = start_time

    @property
    def end_time(self) -> datetime:
        """
        end time within modeled system
        """

        return self.__end_time

    @end_time.setter
    def end_time(self, end_time: datetime):
        end_time = parse_datetime(end_time)
        self.__end_time = end_time
        if self.end_time < self.start_time:
            self.end_time = self.start_time
            self.start_time = end_time

    @property
    def duration(self) -> timedelta:
        """
        duration of run within modeled system
        """

        return self.end_time - self.start_time

    @property
    def interval(self) -> timedelta:
        """
        run sequence interval within the modeled system
        """

        return self.__sequence.interval

    @interval.setter
    def interval(self, interval: timedelta):
        self.__sequence.interval = interval

    @property
    def attributes(self):
        return self.__sequence.attributes

    @attributes.setter
    def attributes(self, attributes: Dict[str, str]):
        self.__sequence.attributes = attributes

    @property
    def models(self) -> List[ModelEntry]:
        """
        models in execution order
        """

        return self.__sequence.models

    @property
    def processors(self) -> int:
        """
        number of PETs / processors / tasks
        """

        return self.__sequence.processors

    @property
    def sequence(self) -> List[str]:
        """
        model execution order
        """

        return [entry.sequence_entry for entry in self.__sequence.sequence]

    @sequence.setter
    def sequence(self, sequence: List[str]):
        sequence_entries = []
        entries = {entry.sequence_entry: entry for entry in self.__sequence.sequence}
        for entry in sequence:
            if entry.upper() in entries:
                sequence_entries.append(entries[entry.upper()])
            elif '->' in entry:
                models = [model.strip() for model in entry.split('->')]
                if len(models) == 2:
                    connection = ConnectionEntry.from_string(entry)
                    source = connection.source.entry_type.value
                    destination = connection.target.entry_type.value
                    del connection

                    for connection in self.__sequence.connections:
                        if isinstance(connection, ConnectionEntry):
                            if (
                                connection.source.entry_type.value == source
                                and connection.target.entry_type.value == destination
                            ):
                                sequence_entries.append(connection)
                                break
                        elif isinstance(connection, MediationEntry):
                            if (connection.sources is None and source is None) or (
                                connection.sources is not None
                                and source
                                in [source.entry_type.value for source in connection.sources]
                            ):
                                if (connection.targets is None and destination is None) or (
                                    connection.targets is not None
                                    and destination
                                    in [
                                        target.entry_type.value
                                        for target in connection.targets
                                    ]
                                ):
                                    sequence_entries.append(connection)
                                    break
                    else:
                        raise KeyError(f'"{entry}" not in {self.connections}')
                elif len(models) == 3:
                    for mediation in self.__sequence.mediations:
                        if models == [
                            model.entry_type.value
                            for model in mediation.models
                            if model is not None
                        ]:
                            sequence_entries.append(mediation)
                            break
                    else:
                        raise KeyError(f'"{entry}" not in {self.connections}')
            else:
                raise KeyError(f'"{entry}" not in {self.sequence}')
        self.__sequence.sequence = sequence_entries

    def connect(self, source: str, target: str = None, method: str = None):
        """
        couple two models with an information exchange pathway

        :param source: model providing information
        :param target: model receiving information
        :param method: remapping method
        """

        if target is None:
            try:
                connection = ConnectionEntry.from_string(source)
                source = connection.source.name
                target = connection.target.name
                del connection
            except:
                pass

        model_types = [model_type.value.upper() for model_type in EntryType]
        remap_methods = [remap.value.lower() for remap in GridRemapMethod]

        if source.upper() not in model_types:
            raise KeyError(f'"{source}" not in {model_types}')
        if target is not None:
            if target.upper() not in model_types:
                raise KeyError(f'"{target}" not in {model_types}')
            target = EntryType(target.upper())
        if method is not None:
            if method.lower() not in remap_methods:
                raise KeyError(f'"{method}" not in {remap_methods}')
            method = GridRemapMethod(method.lower())

        self.__sequence.connect(EntryType(source.upper()), target, method)

    @property
    def connections(self) -> List[str]:
        """
        string representations of coupling connections in format ``'WAV -> HYD'``
        """

        return [str(connection) for connection in self.__sequence.connections]

    def mediate(
        self,
        sources: List[str] = None,
        functions: List[str] = None,
        targets: List[str] = None,
        method: GridRemapMethod = None,
        processors: int = None,
        **attributes,
    ):
        """
        create a mediation between one or two models and a mediator,
        with an arbitrary number of mediation functions

        :param sources: model providing information
        :param functions: mediator functions
        :param targets: model receiving information
        :param method: remapping method
        :param processors: number of processors to assign to mediation
        """

        if targets is None:
            try:
                targets = MediationEntry.from_string(sources).targets
            except:
                pass

        model_types = [model_type.value.upper() for model_type in EntryType]
        remap_methods = [remap.value.lower() for remap in GridRemapMethod]

        if sources is not None:
            if isinstance(sources, str):
                sources = [sources]
            for index, source in enumerate(sources):
                if isinstance(source, str):
                    if source.upper() not in model_types:
                        raise KeyError(f'"{source}" not in {model_types}')
                    sources[index] = EntryType(source.upper())
        if targets is not None:
            if isinstance(targets, str):
                targets = [targets]
            for index, target in enumerate(targets):
                if isinstance(target, str):
                    if target.upper() not in model_types:
                        raise KeyError(f'"{target}" not in {model_types}')
                    targets[index] = EntryType(target.upper())
        if method is not None and isinstance(method, str):
            if method.lower() not in remap_methods:
                raise KeyError(f'"{method}" not in {remap_methods}')
            method = GridRemapMethod(method.lower())

        self.__sequence.mediate(sources, functions, targets, method, processors, **attributes)

    @property
    def __configuration_files(self) -> List[ConfigurationFile]:
        return [
            NEMSConfigurationFile(self.__sequence),
            FileForcingsFile(self.__sequence),
            ModelConfigurationFile(self.start_time, self.duration, self.__sequence),
        ]

    @property
    def configuration(self) -> Dict[str, str]:
        return {
            configuration_file.name: str(configuration_file)
            for configuration_file in self.__configuration_files
        }

    def write(
        self, directory: PathLike, overwrite: bool = False, include_version: bool = False,
    ) -> List[Path]:
        """
        write NEMS / NUOPC configuration to the given directory

        :param directory: path to output directory
        :param overwrite: overwrite existing files
        :param include_version: include the NEMSpy version in a comment
        :returns: list of written file paths
        """

        directory = ensure_directory(directory)
        filenames = []
        for configuration_file in self.__configuration_files:
            filename = configuration_file.write(
                directory / configuration_file.name, overwrite, include_version
            )
            filenames.append(filename)
            if isinstance(configuration_file, ModelConfigurationFile):
                filenames.append(directory / 'atm_namelist.rc')
        return filenames

    def __getitem__(self, model_type: str) -> ModelEntry:
        if not isinstance(model_type, str) and not isinstance(model_type, EntryType):
            raise ValueError(
                f'model type must be {str} or {EntryType}, not {type(model_type)}'
            )
        model_types = [model_type.value.upper() for model_type in EntryType]
        if model_type.upper() not in model_types:
            raise KeyError(f'"{model_type}" not in {model_types}')
        return self.__sequence[EntryType(model_type.upper())]

    def __setitem__(self, model_type: str, model: ModelEntry):
        if not isinstance(model_type, str) and not isinstance(model_type, EntryType):
            raise ValueError(
                f'model type must be {str} or {EntryType}, not {type(model_type)}'
            )
        model_types = [model_type.value.upper() for model_type in EntryType]
        if model_type.upper() not in model_types:
            raise KeyError(f'"{model_type}" not in {model_types}')
        self.__sequence[EntryType(model_type.upper())] = model

    def __contains__(self, model_type: str) -> bool:
        if not isinstance(model_type, str) and not isinstance(model_type, EntryType):
            raise ValueError(
                f'model type must be {str} or {EntryType}, not {type(model_type)}'
            )
        return EntryType(model_type.upper()) in self.__sequence

    def __repr__(self) -> str:
        models = [f'{model.entry_type}={repr(model)}' for model in self.__sequence.models]
        return f'{self.__class__.__name__}({repr(self.interval)}, {", ".join(models)})'



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from enum import Enum
from importlib import metadata as importlib_metadata
import os
from os import PathLike
from pathlib import Path
from textwrap import indent
from typing import Iterator, List, Tuple, Union

from .model.base import (
    AttributeEntry,
    ConnectionEntry,
    EntryType,
    FileForcingEntry,
    GridRemapMethod,
    INDENTATION,
    MediationEntry,
    MediatorEntry,
    ModelEntry,
    SequenceEntry,
    VerbosityOption,
)
from .utilities import create_symlink, LOGGER


class Earth(AttributeEntry):
    """
    multi-model coupling container representing the entire Earth system

    Only one of each model type can be assigned to the Earth system model at a time.
    """

    entry_title = 'EARTH'

    def __init__(self, **models):
        """
        :param atm: atmospheric wind model
        :param wav: oceanic wave model
        :param ocn: oceanic circulation model
        :param hyd: terrestrial water model
        :param med: model mediator
        """

        if 'Verbosity' not in models:
            models['Verbosity'] = VerbosityOption.OFF

        self.__models = {model_type: None for model_type in EntryType}

        attributes = {}
        for key, value in models.items():
            if key.upper() in {entry.name for entry in EntryType}:
                if isinstance(value, ModelEntry):
                    self[EntryType[key.upper()]] = value
            else:
                attributes[key] = value
        self.attributes = attributes

    @property
    def models(self):
        """
        list of models comprising the Earth system
        """

        return self.__models

    def __getitem__(self, model_type: EntryType) -> ModelEntry:
        return self.__models[model_type]

    def __setitem__(self, model_type: EntryType, model: ModelEntry):
        assert model_type == model.entry_type
        if self.__models[model_type] is not None:
            LOGGER.warning(
                f'overwriting existing "{model_type.name}" model: ' f'{repr(self[model_type])}'
            )
        self.__models[model_type] = model

    def __contains__(self, model_type: EntryType):
        return model_type in self.__models

    def __iter__(self) -> Iterator[Tuple[EntryType, ModelEntry]]:
        for model_type, model in self.models.items():
            yield model_type, model

    def __str__(self) -> str:
        attributes = [
            f'{attribute} = {value if not isinstance(value, Enum) else value.value}'
            for attribute, value in self.attributes.items()
        ]

        return '\n'.join(
            [
                f'{self.entry_title}_component_list: '
                f'{" ".join(model_type.value for model_type, model in self.models.items() if model is not None)}',
                f'{self.entry_title}_attributes::',
                indent('\n'.join(attributes), INDENTATION),
                '::',
            ]
        )

    def __repr__(self) -> str:
        models = [
            f'{model_type.name}={repr(model)}' for model_type, model in self.models.items()
        ]
        models += [f'{key}={value}' for key, value in self.attributes.items()]
        return (
            f'{self.__class__.__name__}({self.attributes["Verbosity"]}, {", ".join(models)})'
        )


class RunSequence(AttributeEntry, SequenceEntry):
    """
    multi-model container for model entries, defining the sequence in which they run within the modeled time loop

    NOTE: Currently, only one loop is supported. Nested loops will be implemented in a future version of NEMSpy.
    """

    entry_title = 'Run Sequence'

    def __init__(self, interval: timedelta, **kwargs):
        """
        :param interval: time interval to repeat the main loop in modeled time
        """

        self.interval = interval

        if 'Verbosity' not in kwargs:
            kwargs['Verbosity'] = VerbosityOption.OFF

        self.__models = {}
        attributes = {}
        for key, value in kwargs.items():
            model_types = [model_type.value for model_type in EntryType]
            if key.upper() in model_types and isinstance(value, ModelEntry):
                self.__models[EntryType(key.upper())] = value
            else:
                attributes[key] = value
        self.attributes = attributes

        self.__sequence = [
            model for model in self.models if model.entry_type != EntryType.MEDIATOR
        ]
        self.__link_models()

    def append(self, entry: SequenceEntry):
        """
        add a sequence entry
        """

        if isinstance(entry, ModelEntry):
            model_type = entry.entry_type
            if model_type in self.__models:
                del self.__models[model_type]
                self[entry.entry_type] = entry
        self.__sequence.append(entry)

    def extend(self, sequence: List[SequenceEntry]):
        """
        add several sequence entries
        """

        for entry in sequence:
            self.append(entry)

    @property
    def sequence(self) -> List[SequenceEntry]:
        """
        list of sequence entries in order, including model entries and connections / mediations
        """

        return self.__sequence

    @sequence.setter
    def sequence(self, sequence: List[SequenceEntry]):
        """
        set the sequence by passing a list of entries in order
        """

        sequence = list(sequence)
        if sequence != self.__sequence:
            mediator = self.mediator
            self.__models = {}
            if mediator is not None:
                self.mediator = mediator
            for entry in sequence:
                if isinstance(entry, ModelEntry):
                    model_type = entry.entry_type
                    if model_type in self.__models:
                        raise TypeError(
                            f'duplicate model type ' f'"{model_type.name}" in given sequence'
                        )
                    self.__models[model_type] = entry
            self.__link_models()
            self.__sequence = sequence

    def connect(
        self, source: EntryType, target: EntryType, method: GridRemapMethod = None, **kwargs,
    ):
        """
        assign a simple connection (not a mediation) between two model entries within the sequence
        """

        if method is None:
            method = GridRemapMethod.REDISTRIBUTE
        if EntryType.MEDIATOR in [source, target] and self.mediator is None:
            self.mediator = MediatorEntry(**kwargs)
        if source not in self.__models:
            raise KeyError(f'no {source.name} model in sequence')
        if target not in self.__models:
            raise KeyError(f'no {target.name} model in sequence')
        self.append(ConnectionEntry(self[source], self[target], method))

    @property
    def connections(self) -> List[Union[ConnectionEntry, MediationEntry]]:
        """
        list of all connections in the sequence
        """

        return [
            entry
            for entry in self.sequence
            if isinstance(entry, ConnectionEntry) or isinstance(entry, MediationEntry)
        ]

    @property
    def mediator(self) -> MediatorEntry:
        """
        shortcut property to the mediator entry
        """

        if EntryType.MEDIATOR in self:
            return self.__models[EntryType.MEDIATOR]
        else:
            return None

    @mediator.setter
    def mediator(self, mediator: MediatorEntry):
        """
        set the mediator entry (does not exist in the sequence by itself)
        """

        self[EntryType.MEDIATOR] = mediator

    def mediate(
        self,
        sources: List[EntryType] = None,
        functions: List[str] = None,
        targets: List[EntryType] = None,
        method: GridRemapMethod = None,
        processors: int = None,
        **attributes,
    ):
        """
        assign a mediation between two entries in the sequence
        """

        if 'name' not in attributes:
            attributes['name'] = 'mediator'
        if self.mediator is None:
            self.mediator = MediatorEntry(processors=processors, **attributes)
        else:
            self.mediator.attributes.update(attributes)
        if processors is not None:
            # increase mediation processor assignment if required
            if self.mediator.processors < processors:
                self.mediator.processors = processors

        if sources is not None:
            sources = [self[source] for source in sources]
        if targets is not None:
            targets = [self[target] for target in targets]

        self.append(MediationEntry(self.mediator, sources, functions, targets, method))

    @property
    def mediations(self) -> List[MediationEntry]:
        """
        list of all mediations in the sequence
        """

        return [entry for entry in self.sequence if isinstance(entry, MediationEntry)]

    @property
    def earth(self) -> Earth:
        """
        Earth system assigned to the sequence
        """

        return Earth(
            **{model.entry_type.name: model for model in self.models}, **self.attributes
        )

    @property
    def processors(self) -> int:
        """
        total number of processors assigned to sequence entries
        """

        return sum(model.processors for model in self.__models.values())

    def __link_models(self):
        """
        link entries and assign processors
        """

        models = self.models
        for model in models:
            if model.previous is not None:
                model.previous.next = None
            if model.next is not None:
                model.next = None
            model.start_processor = 0
        for model_index, model in enumerate(models):
            previous_model_index = model_index - 1
            if previous_model_index >= 0:
                model.previous = models[previous_model_index]

    def __setitem__(self, model_type: EntryType, model: ModelEntry):
        assert model_type == model.entry_type
        if model_type in self.__models:
            existing_model = self.__models[model_type]
            LOGGER.warning(
                f'overwriting {model_type.name} model ' f'"{existing_model}" with "{model}"'
            )
            self.__sequence.remove(self.__sequence.index(existing_model))
        self.__models[model_type] = model
        self.__link_models()

    def __getitem__(self, model_type: EntryType) -> ModelEntry:
        return self.__models[model_type]

    @property
    def models(self) -> List[ModelEntry]:
        """
        list of models in the run sequence
        """

        models = [
            model
            for model_type, model in self.__models.items()
            if model_type in self and model_type is not EntryType.MEDIATOR
        ]
        if self.mediator is not None:
            models.insert(0, self.mediator)
        return models

    def __iter__(self) -> Iterator[ModelEntry]:
        for model in self.models:
            yield model

    def __contains__(self, model_type: EntryType) -> bool:
        return model_type in self.__models

    def __len__(self) -> int:
        return len(self.sequence)

    @property
    def sequence_entry(self) -> str:
        return str(self)

    def __str__(self) -> str:
        block = '\n'.join(
            [
                f'@{self.interval / timedelta(seconds=1):.0f}',
                indent(
                    '\n'.join(entry.sequence_entry for entry in self.__sequence), INDENTATION
                ),
                '@',
            ]
        )
        return '\n'.join([f'runSeq::', indent(block, INDENTATION), '::'])

    def __repr__(self) -> str:
        models = [f'{model.entry_type.name.lower()}={repr(model)}' for model in self.models]
        return f'{self.__class__.__name__}({repr(self.interval)}, {", ".join(models)})'


class ConfigurationFile(ABC):
    """
    abstraction of a configuration file
    """

    name: str = NotImplementedError

    def __init__(self, sequence: RunSequence):
        """
        :param sequence: run sequence object containing models and order
        """

        self.sequence = sequence

    def __getitem__(self, entry_type: type) -> List[AttributeEntry]:
        return [entry for entry in self if isinstance(entry, entry_type)]

    @property
    def version_header(self) -> str:
        """
        comment header indicating filename and NEMSpy version
        """

        installed_distributions = importlib_metadata.distributions()
        for distribution in installed_distributions:
            if distribution.metadata['Name'].lower() == 'nemspy':
                version = distribution.version
                break
        else:
            version = 'unknown'
        return f'# `{self.name}` generated with NEMSpy {version}'

    def write(
        self, filename: PathLike, overwrite: bool = False, include_version: bool = False
    ) -> Path:
        """
        write this configuration to file

        :param filename: path to file
        :param overwrite: overwrite an existing file
        :param include_version: include NEMSpy version information
        :returns: path to written file
        """

        if not isinstance(filename, Path):
            filename = Path(filename)
        ensure_directory(filename.parent)

        output = f'{self}\n'
        if include_version:
            output = f'{self.version_header}\n' f'{output}'

        if filename.is_dir():
            filename = filename / self.name
            LOGGER.warning(
                f'creating new file "{os.path.relpath(filename.resolve(), Path.cwd())}"'
            )

        if filename.exists():
            LOGGER.warning(
                f'{"overwriting" if overwrite else "skipping"} existing file "{os.path.relpath(filename.resolve(), Path.cwd())}"'
            )
        if not filename.exists() or overwrite:
            with open(filename, 'w', newline='\n') as output_file:
                output_file.write(output)

        return filename

    @abstractmethod
    def __str__(self) -> str:
        """
        :returns: string representation of configuration file
        """

        raise NotImplementedError

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.sequence)})'


class NEMSConfigurationFile(ConfigurationFile):
    """
    ``nems.configure`` file, containing NEMS members, coupling connections, and run sequence information
    """

    name = 'nems.configure'

    @property
    def entries(self) -> List[AttributeEntry]:
        """
        list of attributed configuration entries, including the Earth system, models, and run sequence
        """

        return [self.sequence.earth, *self.sequence.models, self.sequence]

    def __iter__(self) -> Iterator[AttributeEntry]:
        for entry in self.entries:
            yield entry

    def __str__(self) -> str:
        return '\n'.join(f'# {entry.entry_title} #\n' f'{entry}\n' for entry in self).strip()


class FileForcingsFile(ConfigurationFile):
    """
    ``config.rc`` file, containing paths to forcing files
    """

    name = 'config.rc'

    @property
    def entries(self) -> List[FileForcingEntry]:
        """
        list of file forcing entries
        """

        return [entry for entry in self.sequence if isinstance(entry, FileForcingEntry)]

    def __iter__(self) -> Iterator[AttributeEntry]:
        for entry in self.entries:
            yield entry

    def __str__(self) -> str:
        return '\n'.join([FileForcingEntry.__str__(model_mesh) for model_mesh in self])


class ModelConfigurationFile(ConfigurationFile):
    """
    ``model_configure`` file, containing information on modeled start and end times, as well as ensemble information also aliased to ``atm_namelist.rc``
    """

    name = 'model_configure'

    def __init__(
        self,
        start_time: datetime,
        duration: timedelta,
        sequence: RunSequence,
        create_atm_namelist_rc: bool = True,
    ):
        """
        :param start_time: start time in model time
        :param duration: duration in model time
        :param sequence: run sequence containing models, connections, and order
        :param create_atm_namelist_rc: whether to create a symlink to ``atm_namelist.rc``
        """

        self.start_time = start_time
        self.duration = duration
        self.create_atm_namelist_rc = create_atm_namelist_rc
        super().__init__(sequence)

    def write(
        self, filename: PathLike, overwrite: bool = False, include_version: bool = False,
    ) -> Path:
        filename = super().write(filename, overwrite, include_version)
        if self.create_atm_namelist_rc:
            create_symlink(filename, filename.parent / 'atm_namelist.rc', relative=True)
        return filename

    def __str__(self) -> str:
        duration_hours = round(self.duration / timedelta(hours=1))
        return '\n'.join(
            [
                'total_member:            1',
                'print_esmf:              .true.',
                f'namelist:                {"atm_namelist.rc" if self.create_atm_namelist_rc else "model_configure"}',
                f'PE_MEMBER01:             {self.sequence.processors}',
                f'start_year:              {self.start_time.year}',
                f'start_month:             {self.start_time.month}',
                f'start_day:               {self.start_time.day}',
                f'start_hour:              {self.start_time.hour}',
                f'start_minute:            {self.start_time.minute}',
                f'start_second:            {self.start_time.second}',
                f'nhours_fcst:             {duration_hours:.0f}',
                'RUN_CONTINUE:            .false.',
                'ENS_SPS:                 .false.',
                # 'dt_atmos:                   @[DT_ATMOS]'
                # 'atm_coupling_interval_sec:  @[coupling_interval_fast_sec]'
                #
                # 'iatm: @[IATM]'
                # 'jatm: @[JATM]'
                #
                # 'cdate0: @[CDATE]'
                # 'nfhout: @[NFHOUT]'
                # 'filename_base: @[FILENAME_BASE]'
            ]
        )


def ensure_directory(directory: PathLike) -> Path:
    """
    ensure that a directory exists

    :param directory: directory path to ensure
    :returns: path to ensured directory
    """

    if not isinstance(directory, Path):
        directory = Path(directory)
    directory = directory.expanduser()
    if directory.is_file():
        directory = directory.parent
    if not directory.exists():
        directory.mkdir(parents=True, exist_ok=True)
    return directory



from os import PathLike

from .base import EntryType, FileForcingEntry, ModelEntry


class AtmosphericModelEntry(ModelEntry):
    """
    abstraction of a generic atmospheric model
    """

    entry_type = EntryType.ATMOSPHERIC

    def __init__(self, processors: int, **kwargs):
        super().__init__(processors, **kwargs)


class AtmosphericForcingEntry(AtmosphericModelEntry, FileForcingEntry):
    """
    file forcing entry for Atmospheric Mesh (ATMesh), which combines data from HRRR, GFS, etc.

    >>> atmospheric_mesh = AtmosphericForcingEntry(filename='wind_atm_fin_ch_time_vec.nc', processors=1)
    """

    name = 'atmesh'

    def __init__(self, filename: PathLike = None, processors: int = None, **kwargs):
        if processors is None:
            processors = 1
        AtmosphericModelEntry.__init__(self, processors, **kwargs)
        FileForcingEntry.__init__(self, self.entry_type, filename)


class HWRFEntry(AtmosphericModelEntry):
    """
    Hurricane Weather Research and Forecasting (HWRF) model
    https://en.wikipedia.org/wiki/Hurricane_Weather_Research_and_Forecasting_Model
    """

    name = 'hwrf'

    def __init__(self, processors: int, **kwargs):
        super().__init__(processors, **kwargs)



from .base import EntryType, ModelEntry


class OceanModelEntry(ModelEntry):
    """
    abstraction of a generic oceanic model
    """

    entry_type = EntryType.OCEAN

    def __init__(self, processors: int, **kwargs):
        super().__init__(processors, **kwargs)


class ADCIRCEntry(OceanModelEntry):
    """
    Advanced Circulation (ADCIRC) model
    http://adcirc.org

    >>> ocean_model = ADCIRCEntry(processors=11, Verbosity='max', DumpFields=False)
    """

    name = 'adcirc'

    def __init__(self, processors: int, **kwargs):
        super().__init__(processors, **kwargs)


class SCHISMEntry(OceanModelEntry):
    """
    Semi-implicit Cross-scale Hydroscience Integrated System Model (SCHISM)
    http://ccrm.vims.edu/schismweb/
    """

    name = 'schism'

    def __init__(self, processors: int, **kwargs):
        super().__init__(processors, **kwargs)



from datetime import datetime
import logging
import os
from os import PathLike
from pathlib import Path
import shutil
import sys
from typing import Union


def repository_root(path: PathLike = None) -> Path:
    """
    find enclosing Git repository

    :param path: path to start at
    :return: path to root of Git repository
    """

    if path is None:
        path = __file__
    if not isinstance(path, Path):
        path = Path(path)
    if path.is_file():
        path = path.parent
    if '.git' in (child.name for child in path.iterdir()) or path == path.parent:
        return path
    else:
        return repository_root(path.parent)


def get_logger(
    name: str,
    log_filename: PathLike = None,
    file_level: int = None,
    console_level: int = None,
    log_format: str = None,
) -> logging.Logger:
    """
    instantiate logger instance

    :param name: name of logger
    :param log_filename: path to log file
    :param file_level: minimum log level to write to log file
    :param console_level: minimum log level to print to console
    :param log_format: logger message format
    :return: instance of a Logger object
    """

    if file_level is None:
        file_level = logging.DEBUG
    if console_level is None:
        console_level = logging.INFO
    logger = logging.getLogger(name)

    # check if logger is already configured
    if logger.level == logging.NOTSET and len(logger.handlers) == 0:
        # check if logger has a parent
        if '.' in name:
            logger.parent = get_logger(name.rsplit('.', 1)[0])
        else:
            # otherwise create a new split-console logger
            logger.setLevel(logging.DEBUG)
            if console_level != logging.NOTSET:
                if console_level <= logging.INFO:

                    class LoggingOutputFilter(logging.Filter):
                        def filter(self, rec):
                            return rec.levelno in (logging.DEBUG, logging.INFO)

                    console_output = logging.StreamHandler(sys.stdout)
                    console_output.setLevel(console_level)
                    console_output.addFilter(LoggingOutputFilter())
                    logger.addHandler(console_output)

                console_errors = logging.StreamHandler(sys.stderr)
                console_errors.setLevel(max((console_level, logging.WARNING)))
                logger.addHandler(console_errors)

    if log_filename is not None:
        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(file_level)
        for existing_file_handler in [
            handler for handler in logger.handlers if type(handler) is logging.FileHandler
        ]:
            logger.removeHandler(existing_file_handler)
        logger.addHandler(file_handler)

    if log_format is None:
        log_format = '[%(asctime)s] %(name)-9s %(levelname)-8s: %(message)s'
    log_formatter = logging.Formatter(log_format)
    for handler in logger.handlers:
        handler.setFormatter(log_formatter)

    return logger


LOGGER = get_logger('nemspy')


def create_symlink(
    source_filename: PathLike, symlink_filename: PathLike, relative: bool = False
):
    """
    :param source_filename: path to point to
    :param symlink_filename: path at which to create symlink
    :param relative: make symlink relative to source location
    """

    if not isinstance(source_filename, Path):
        source_filename = Path(source_filename)
    if not isinstance(symlink_filename, Path):
        symlink_filename = Path(symlink_filename)

    if symlink_filename.is_symlink():
        LOGGER.debug(f'removing symlink "{symlink_filename}"')
        os.remove(symlink_filename)
    symlink_filename = symlink_filename.parent.absolute().resolve() / symlink_filename.name

    starting_directory = None
    if relative:
        starting_directory = Path().cwd().resolve()
        os.chdir(symlink_filename.parent)
        if source_filename.is_absolute():
            try:
                source_filename = source_filename.relative_to(symlink_filename.parent)
            except ValueError as error:
                LOGGER.warning(error)
                os.chdir(starting_directory)
    else:
        source_filename = source_filename.absolute()

    try:
        symlink_filename.symlink_to(source_filename)
    except Exception as error:
        LOGGER.warning(f'could not create symbolic link: {error}')
        shutil.copyfile(source_filename, symlink_filename)

    if starting_directory is not None:
        os.chdir(starting_directory)


def parse_datetime(value: Union[int, float, str, datetime]) -> datetime:
    """
    attempt to parse a datetime from the given value

    :param value: value to parse
    :return: parsed datetime
    """

    if not isinstance(value, datetime):
        if isinstance(value, int) or isinstance(value, float):
            value = datetime.fromtimestamp(value)
        elif isinstance(value, str):
            value = datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
        else:
            raise TypeError(
                f'unable to convert value of type "{type(value)}" to datetime: {value}'
            )
    return value



from nemspy.model import ADCIRCEntry, AtmosphericForcingEntry, WaveWatch3ForcingEntry
from nemspy.utilities import repository_root

REFERENCE_DIRECTORY = repository_root() / 'tests/reference'
ATMOSPHERIC_MESH_FILENAME = REFERENCE_DIRECTORY / 'wind_atm_fin_ch_time_vec.nc'
WAVE_MESH_FILENAME = REFERENCE_DIRECTORY / 'ww3.Constant.20151214_sxy_ike_date.nc'


def test_model():
    model = AtmosphericForcingEntry(
        ATMOSPHERIC_MESH_FILENAME, Verbosity='off', test='value', test2=5,
    )
    model.start_processor = 0

    assert (
        str(model) == 'ATM_model:                      atmesh\n'
        'ATM_petlist_bounds:             0 0\n'
        'ATM_attributes::\n'
        '  Verbosity = off\n'
        '  test = value\n'
        '  test2 = 5\n'
        '::'
    )


def test_processors():
    model_1 = AtmosphericForcingEntry(ATMOSPHERIC_MESH_FILENAME)
    model_2 = WaveWatch3ForcingEntry(WAVE_MESH_FILENAME)
    model_3 = ADCIRCEntry(11)

    model_1.next = model_2
    model_2.next = model_3

    assert model_1.start_processor is None
    assert model_1.end_processor is None
    assert model_2.start_processor is None
    assert model_2.end_processor is None
    assert model_3.start_processor is None
    assert model_3.end_processor is None

    model_1.start_processor = 0

    assert model_1.start_processor == 0
    assert model_1.end_processor == 0
    assert model_2.start_processor == 1
    assert model_2.end_processor == 1
    assert model_3.start_processor == 2
    assert model_3.end_processor == 12

    model_2.processors = 3
    model_1.processors = 4

    assert model_1.start_processor == 0
    assert model_1.end_processor == 3
    assert model_2.start_processor == 4
    assert model_2.end_processor == 6
    assert model_3.start_processor == 7
    assert model_3.end_processor == 17


def test_from_string():
    model = AtmosphericForcingEntry.from_string(
        'ATM_model:                      atmesh\n'
        'ATM_petlist_bounds:             0 0\n'
        'ATM_attributes::\n'
        '  Verbosity = off\n'
        '  test = value\n'
        '  test2 = 5\n'
        '::',
        filename='wind.nc',
    )

    assert isinstance(model, AtmosphericForcingEntry)

    assert model.name == 'atmesh'
    assert model.processors == 1
    assert model.attributes == {
        'Verbosity': 'off',
        'test': 'value',
        'test2': '5',
    }
